from __future__ import annotations

from dataclasses import dataclass

import mujoco
import numpy as np

from ..exceptions import InvalidMocapBody
from .base import MatrixLieGroup
from .so3 import SO3
from .utils import get_epsilon, skew

_IDENTITY_WXYZ_XYZ = np.array([1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], dtype=np.float64)


@dataclass(frozen=True)
class SE3(MatrixLieGroup):
    """Special Euclidean group for proper rigid transforms in 3D.

    Internal parameterization is (qw, qx, qy, qz, x, y, z). Tangent parameterization is
    (vx, vy, vz, omega_x, omega_y, omega_z).
    """

    wxyz_xyz: np.ndarray
    matrix_dim: int = 4
    parameters_dim: int = 7
    tangent_dim: int = 6
    space_dim: int = 3

    def __repr__(self) -> str:
        quat = np.round(self.wxyz_xyz[:4], 5)
        xyz = np.round(self.wxyz_xyz[4:], 5)
        return f"{self.__class__.__name__}(wxyz={quat}, xyz={xyz})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SE3):
            return NotImplemented
        return np.array_equal(self.wxyz_xyz, other.wxyz_xyz)

    def copy(self) -> SE3:
        return SE3(wxyz_xyz=np.array(self.wxyz_xyz))

    def parameters(self) -> np.ndarray:
        return self.wxyz_xyz

    @classmethod
    def identity(cls) -> SE3:
        return SE3(wxyz_xyz=_IDENTITY_WXYZ_XYZ)

    @classmethod
    def from_rotation_and_translation(
        cls,
        rotation: SO3,
        translation: np.ndarray,
    ) -> SE3:
        assert translation.shape == (SE3.space_dim,)
        return SE3(wxyz_xyz=np.concatenate([rotation.wxyz, translation]))

    @classmethod
    def from_rotation(cls, rotation: SO3) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=rotation, translation=np.zeros(SE3.space_dim)
        )

    @classmethod
    def from_translation(cls, translation: np.ndarray) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3.identity(), translation=translation
        )

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> SE3:
        assert matrix.shape == (SE3.matrix_dim, SE3.matrix_dim)
        return SE3.from_rotation_and_translation(
            rotation=SO3.from_matrix(matrix[:3, :3]),
            translation=matrix[:3, 3],
        )

    @classmethod
    def from_mocap_id(cls, data: mujoco.MjData, mocap_id: int) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3(data.mocap_quat[mocap_id]),
            translation=data.mocap_pos[mocap_id],
        )

    @classmethod
    def from_mocap_name(
        cls, model: mujoco.MjModel, data: mujoco.MjData, mocap_name: str
    ) -> SE3:
        mocap_id = model.body(mocap_name).mocapid[0]
        if mocap_id == -1:
            raise InvalidMocapBody(mocap_name, model)
        return SE3.from_mocap_id(data, mocap_id)

    @classmethod
    def sample_uniform(cls) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3.sample_uniform(),
            translation=np.random.uniform(-1.0, 1.0, size=(SE3.space_dim,)),
        )

    def rotation(self) -> SO3:
        return SO3(wxyz=self.wxyz_xyz[:4])

    def translation(self) -> np.ndarray:
        return self.wxyz_xyz[4:]

    def as_matrix(self) -> np.ndarray:
        hmat = np.eye(self.matrix_dim, dtype=np.float64)
        hmat[:3, :3] = self.rotation().as_matrix()
        hmat[:3, 3] = self.translation()
        return hmat

    @classmethod
    def exp(cls, tangent: np.ndarray) -> SE3:
        assert tangent.shape == (cls.tangent_dim,)
        rotation = SO3.exp(tangent[3:])
        theta = np.float64(mujoco.mju_norm3(tangent[3:]))
        t2 = theta * theta
        if t2 < get_epsilon(t2.dtype):
            v_mat = rotation.as_matrix()
        else:
            skew_omega = skew(tangent[3:])
            v_mat = (
                np.eye(3, dtype=np.float64)
                + (1.0 - np.cos(theta)) / t2 * skew_omega
                + (theta - np.sin(theta)) / (t2 * theta) * (skew_omega @ skew_omega)
            )
        return cls.from_rotation_and_translation(
            rotation=rotation,
            translation=v_mat @ tangent[:3],
        )

    def inverse(self) -> SE3:
        inverse_wxyz_xyz = np.empty(SE3.parameters_dim, dtype=np.float64)
        mujoco.mju_negQuat(inverse_wxyz_xyz[:4], self.wxyz_xyz[:4])
        mujoco.mju_rotVecQuat(
            inverse_wxyz_xyz[4:], -1.0 * self.wxyz_xyz[4:], inverse_wxyz_xyz[:4]
        )
        return SE3(wxyz_xyz=inverse_wxyz_xyz)

    def normalize(self) -> SE3:
        normalized_wxyz_xyz = np.array(self.wxyz_xyz)
        mujoco.mju_normalize4(normalized_wxyz_xyz[:4])
        return SE3(wxyz_xyz=normalized_wxyz_xyz)

    def apply(self, target: np.ndarray) -> np.ndarray:
        assert target.shape == (SE3.space_dim,)
        rotated_target = np.empty(SE3.space_dim, dtype=np.float64)
        mujoco.mju_rotVecQuat(rotated_target, target, self.wxyz_xyz[:4])
        return rotated_target + self.wxyz_xyz[4:]

    def multiply(self, other: SE3) -> SE3:
        wxyz_xyz = np.empty(SE3.parameters_dim, dtype=np.float64)
        mujoco.mju_mulQuat(wxyz_xyz[:4], self.wxyz_xyz[:4], other.wxyz_xyz[:4])
        mujoco.mju_rotVecQuat(wxyz_xyz[4:], other.wxyz_xyz[4:], self.wxyz_xyz[:4])
        wxyz_xyz[4:] += self.wxyz_xyz[4:]
        return SE3(wxyz_xyz=wxyz_xyz)

    def log(self) -> np.ndarray:
        omega = self.rotation().log()
        theta = np.float64(mujoco.mju_norm3(omega))
        t2 = theta * theta
        skew_omega = skew(omega)
        skew_omega2 = skew_omega @ skew_omega
        if t2 < get_epsilon(t2.dtype):
            vinv_mat = (
                np.eye(3, dtype=np.float64) - 0.5 * skew_omega + skew_omega2 / 12.0
            )
        else:
            half_theta = 0.5 * theta
            vinv_mat = (
                np.eye(3, dtype=np.float64)
                - 0.5 * skew_omega
                + (1.0 - 0.5 * theta * np.cos(half_theta) / np.sin(half_theta))
                / t2
                * skew_omega2
            )
        tangent = np.empty(SE3.tangent_dim, dtype=np.float64)
        tangent[:3] = vinv_mat @ self.translation()
        tangent[3:] = omega
        return tangent

    def adjoint(self) -> np.ndarray:
        rotation = self.rotation()
        rotation_mat = rotation.as_matrix()
        tangent_mat = skew(self.translation()) @ rotation_mat
        adjoint_mat = np.zeros((SE3.tangent_dim, SE3.tangent_dim), dtype=np.float64)
        adjoint_mat[:3, :3] = rotation_mat
        adjoint_mat[:3, 3:] = tangent_mat
        adjoint_mat[3:, 3:] = rotation_mat
        return adjoint_mat

    # Jacobians.

    # Eqn 179 a)
    @classmethod
    def ljac(cls, other: np.ndarray) -> np.ndarray:
        theta_squared = np.float64(mujoco.mju_dot3(other[3:], other[3:]))
        if theta_squared < get_epsilon(theta_squared.dtype):
            return np.eye(cls.tangent_dim)
        ljac_se3 = np.zeros((cls.tangent_dim, cls.tangent_dim), dtype=np.float64)
        ljac_translation = _getQ(other)
        ljac_so3 = SO3.ljac(other[3:])
        ljac_se3[:3, :3] = ljac_so3
        ljac_se3[:3, 3:] = ljac_translation
        ljac_se3[3:, 3:] = ljac_so3
        return ljac_se3

    # Eqn 179 b)
    @classmethod
    def ljacinv(cls, other: np.ndarray) -> np.ndarray:
        theta_squared = np.float64(mujoco.mju_dot3(other[3:], other[3:]))
        if theta_squared < get_epsilon(theta_squared.dtype):
            return np.eye(cls.tangent_dim)
        ljacinv_se3 = np.zeros((cls.tangent_dim, cls.tangent_dim), dtype=np.float64)
        ljac_translation = _getQ(other)
        ljacinv_so3 = SO3.ljacinv(other[3:])
        ljacinv_se3[:3, :3] = ljacinv_so3
        ljacinv_se3[:3, 3:] = -ljacinv_so3 @ ljac_translation @ ljacinv_so3
        ljacinv_se3[3:, 3:] = ljacinv_so3
        return ljacinv_se3


# Eqn 180.
def _getQ(c) -> np.ndarray:
    theta = np.float64(mujoco.mju_norm3(c[3:]))
    t2 = theta * theta
    A = 0.5
    if t2 < get_epsilon(t2.dtype):
        B = (1.0 / 6.0) + (1.0 / 120.0) * t2
        C = -(1.0 / 24.0) + (1.0 / 720.0) * t2
        D = -(1.0 / 60.0)
    else:
        t4 = t2 * t2
        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)
        B = (theta - sin_theta) / (t2 * theta)
        C = (1.0 - 0.5 * t2 - cos_theta) / t4
        D = (2.0 * theta - 3.0 * sin_theta + theta * cos_theta) / (2.0 * t4 * theta)
    V = skew(c[:3])
    W = skew(c[3:])
    VW = V @ W
    WV = VW.T
    WVW = WV @ W
    VWW = VW @ W
    return (
        +A * V
        + B * (WV + VW + WVW)
        - C * (VWW - VWW.T - 3.0 * WVW)
        + D * (WVW @ W + W @ WVW)
    )



"""
This file contains key codes for keyboard input handling.
"""

KEY_UNKNOWN = -1
KEY_SPACE = 32
KEY_APOSTROPHE = 39
KEY_COMMA = 44
KEY_MINUS = 45
KEY_PERIOD = 46
KEY_SLASH = 47
KEY_0 = 48
KEY_1 = 49
KEY_2 = 50
KEY_3 = 51
KEY_4 = 52
KEY_5 = 53
KEY_6 = 54
KEY_7 = 55
KEY_8 = 56
KEY_9 = 57
KEY_SEMICOLON = 59
KEY_EQUAL = 61
KEY_A = 65
KEY_B = 66
KEY_C = 67
KEY_D = 68
KEY_E = 69
KEY_F = 70
KEY_G = 71
KEY_H = 72
KEY_I = 73
KEY_J = 74
KEY_K = 75
KEY_L = 76
KEY_M = 77
KEY_N = 78
KEY_O = 79
KEY_P = 80
KEY_Q = 81
KEY_R = 82
KEY_S = 83
KEY_T = 84
KEY_U = 85
KEY_V = 86
KEY_W = 87
KEY_X = 88
KEY_Y = 89
KEY_Z = 90
KEY_LEFT_BRACKET = 91
KEY_BACKSLASH = 92
KEY_RIGHT_BRACKET = 93
KEY_GRAVE_ACCENT = 96
KEY_ESCAPE = 256
KEY_ENTER = 257
KEY_TAB = 258
KEY_BACKSPACE = 259
KEY_INSERT = 260
KEY_DELETE = 261
KEY_RIGHT = 262
KEY_LEFT = 263
KEY_DOWN = 264
KEY_UP = 265
KEY_PAGE_UP = 266
KEY_PAGE_DOWN = 267
KEY_HOME = 268
KEY_END = 269
KEY_CAPS_LOCK = 280
KEY_SCROLL_LOCK = 281
KEY_NUM_LOCK = 282
KEY_PRINT_SCREEN = 283
KEY_PAUSE = 284
KEY_F1 = 290
KEY_F2 = 291
KEY_F3 = 292
KEY_F4 = 293
KEY_F5 = 294
KEY_F6 = 295
KEY_F7 = 296
KEY_F8 = 297
KEY_F9 = 298
KEY_F10 = 299
KEY_F11 = 300
KEY_F12 = 301
KEY_KP_0 = 320
KEY_KP_1 = 321
KEY_KP_2 = 322
KEY_KP_3 = 323
KEY_KP_4 = 324
KEY_KP_5 = 325
KEY_KP_6 = 326
KEY_KP_7 = 327
KEY_KP_8 = 328
KEY_KP_9 = 329
KEY_KP_DECIMAL = 330
KEY_KP_DIVIDE = 331
KEY_KP_MULTIPLY = 332
KEY_KP_SUBTRACT = 333
KEY_KP_ADD = 334
KEY_KP_ENTER = 335
KEY_KP_EQUAL = 336
KEY_LEFT_SHIFT = 340
KEY_LEFT_CONTROL = 341
KEY_LEFT_ALT = 342
KEY_LEFT_SUPER = 343
KEY_RIGHT_SHIFT = 344
KEY_RIGHT_CONTROL = 345
KEY_RIGHT_ALT = 346
KEY_RIGHT_SUPER = 347



"""Equality constraint task implementation."""

from __future__ import annotations

import logging
from typing import List, Optional, Sequence

import mujoco
import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from .exceptions import InvalidConstraint, TaskDefinitionError
from .task import Task


def _get_constraint_dim(constraint: int) -> int:
    """Return the dimension of an equality constraint in the efc* arrays."""
    return {
        mujoco.mjtEq.mjEQ_CONNECT.value: 3,
        mujoco.mjtEq.mjEQ_WELD.value: 6,
        mujoco.mjtEq.mjEQ_JOINT.value: 1,
        mujoco.mjtEq.mjEQ_TENDON.value: 1,
    }[constraint]


def _get_dense_constraint_jacobian(
    model: mujoco.MjModel, data: mujoco.MjData
) -> np.ndarray:
    """Return the dense constraint Jacobian for a model."""
    if mujoco.mj_isSparse(model):
        efc_J = np.empty((data.nefc, model.nv))
        mujoco.mju_sparse2dense(
            efc_J,
            data.efc_J,
            data.efc_J_rownnz,
            data.efc_J_rowadr,
            data.efc_J_colind,
        )
        return efc_J
    return data.efc_J.reshape((data.nefc, model.nv)).copy()


class EqualityConstraintTask(Task):
    """Regulate equality constraints in a model.

    Equality constraints are useful, among other things, for modeling "loop joints"
    such as four-bar linkages. In MuJoCo, there are several types of equality
    constraints, including:

    * ``mjEQ_CONNECT``: Connect two bodies at a point (ball joint).
    * ``mjEQ_WELD``: Fix relative pose of two bodies.
    * ``mjEQ_JOINT``: Couple the values of two scalar joints.
    * ``mjEQ_TENDON``: Couple the values of two tendons.

    This task can regulate all equality constraints in a model or a specific subset
    identified by name or ID.

    Attributes:
        equalities: ID or name of the equality constraints to regulate. If not provided,
            the task will regulate all equality constraints in the model.
        cost: Cost vector for the equality constraint task. Either a scalar, in which
            case the same cost is applied to all constraints, or a vector of shape
            ``(neq,)``, where ``neq`` is the number of equality constraints in the
            model.

    Raises:
        InvalidConstraint: If a specified equality constraint name or ID is not found,
            or if the constraint is not active at the initial configuration.
        TaskDefinitionError: If no equality constraints are found or if cost parameters
            have invalid shape or values.

    Example:

    .. code-block:: python

        # Regulate all equality constraints with the same cost.
        eq_task = EqualityConstraintTask(model, cost=1.0)

        # Regulate specific equality constraints with different costs.
        eq_task = EqualityConstraintTask(
            model,
            cost=[1.0, 0.5],
            equalities=["connect_right", "connect_left"]
        )
    """

    def __init__(
        self,
        model: mujoco.MjModel,
        cost: npt.ArrayLike,
        equalities: Optional[Sequence[int | str]] = None,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        self._eq_ids = self._resolve_equality_ids(model, equalities)
        self._eq_types = model.eq_type[self._eq_ids].copy()
        self._neq_total = len(self._eq_ids)
        self._mask: np.ndarray | None = None

        super().__init__(cost=np.zeros((1,)), gain=gain, lm_damping=lm_damping)
        self.set_cost(cost)

    def set_cost(self, cost: npt.ArrayLike) -> None:
        """Set the cost vector for the equality constraint task.

        Args:
            cost: Cost vector for the equality constraint task.
        """
        cost = np.atleast_1d(cost)
        if cost.ndim != 1 or cost.shape[0] not in (1, self._neq_total):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} cost must be a vector of shape (1,) "
                f"or ({self._neq_total},). Got {cost.shape}."
            )
        if not np.all(cost >= 0.0):
            raise TaskDefinitionError(f"{self.__class__.__name__} cost must be >= 0")

        # Per constraint cost.
        self._cost = (
            np.full((self._neq_total,), cost[0]) if cost.shape[0] == 1 else cost.copy()
        )

        # Expanded per constraint dimension.
        repeats = [_get_constraint_dim(eq_type) for eq_type in self._eq_types]
        self.cost = np.repeat(self._cost, repeats)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        """Compute the equality constraint task error.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Equality constraint task error vector :math:`e(q)`. The shape of the
            error vector is ``(neq_active * constraint_dim,)``, where ``neq_active``
            is the number of active equality constraints, and ``constraint_dim``
            depends on the type of equality constraint.
        """
        self._update_active_constraints(configuration)
        return configuration.data.efc_pos[self._mask]

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        """Compute the task Jacobian at a given configuration.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Equality constraint task jacobian :math:`J(q)`. The shape of the Jacobian
            is ``(neq_active * constraint_dim, nv)``, where ``neq_active`` is the
            number of active equality constraints, ``constraint_dim`` depends on the
            type of equality constraint, and ``nv`` is the dimension of the tangent
            space.
        """
        self._update_active_constraints(configuration)
        efc_J = _get_dense_constraint_jacobian(configuration.model, configuration.data)
        return efc_J[self._mask]

    # Helper functions.

    def _update_active_constraints(self, configuration: Configuration) -> None:
        self._mask = (
            configuration.data.efc_type == mujoco.mjtConstraint.mjCNSTR_EQUALITY
        ) & np.isin(configuration.data.efc_id, self._eq_ids)
        active_eq_ids = configuration.data.efc_id[self._mask]
        self.cost = self._cost[active_eq_ids]

    def _resolve_equality_ids(
        self, model: mujoco.MjModel, equalities: Optional[Sequence[int | str]]
    ) -> np.ndarray:
        eq_ids: List[int] = []

        if equalities is not None:
            for eq_id_or_name in equalities:
                eq_id: int
                if isinstance(eq_id_or_name, str):
                    eq_id = mujoco.mj_name2id(
                        model, mujoco.mjtObj.mjOBJ_EQUALITY, eq_id_or_name
                    )
                    if eq_id == -1:
                        raise InvalidConstraint(
                            f"Equality constraint '{eq_id_or_name}' not found."
                        )
                else:
                    eq_id = eq_id_or_name
                    if eq_id < 0 or eq_id >= model.neq:
                        raise InvalidConstraint(
                            f"Equality constraint index {eq_id} out of range. "
                            f"Must be in range [0, {model.neq})."
                        )
                if not model.eq_active0[eq_id]:
                    raise InvalidConstraint(
                        f"Equality constraint {eq_id} is not active at initial "
                        "configuration."
                    )
                else:
                    eq_ids.append(eq_id)
            # Check for duplicates.
            if len(eq_ids) != len(set(eq_ids)):
                raise TaskDefinitionError(
                    f"Duplicate equality constraint IDs provided: {eq_ids}."
                )
        else:
            eq_ids = list(range(model.neq))
            logging.info("Regulating %d equality constraints", len(eq_ids))

        # Ensure we have at least 1 constraint.
        if len(eq_ids) == 0:
            raise TaskDefinitionError(
                f"{self.__class__.__name__} no equality constraints found in this "
                "model."
            )

        return np.array(eq_ids)



from __future__ import annotations

from dataclasses import dataclass
from typing import NamedTuple

import mujoco
import numpy as np

from .base import MatrixLieGroup
from .utils import get_epsilon

_IDENTITIY_WXYZ = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)


class RollPitchYaw(NamedTuple):
    """Struct containing roll, pitch, and yaw Euler angles."""

    roll: float
    pitch: float
    yaw: float


@dataclass(frozen=True)
class SO3(MatrixLieGroup):
    """Special orthogonal group for 3D rotations.

    Internal parameterization is (qw, qx, qy, qz). Tangent parameterization is
    (omega_x, omega_y, omega_z).
    """

    wxyz: np.ndarray
    matrix_dim: int = 3
    parameters_dim: int = 4
    tangent_dim: int = 3
    space_dim: int = 3

    def __post_init__(self) -> None:
        if self.wxyz.shape != (self.parameters_dim,):
            raise ValueError(
                f"Expeced wxyz to be a length 4 vector but got {self.wxyz.shape[0]}."
            )

    def __repr__(self) -> str:
        wxyz = np.round(self.wxyz, 5)
        return f"{self.__class__.__name__}(wxyz={wxyz})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SO3):
            return NotImplemented
        return np.array_equal(self.wxyz, other.wxyz)

    def parameters(self) -> np.ndarray:
        return self.wxyz

    def copy(self) -> SO3:
        return SO3(wxyz=self.wxyz.copy())

    @classmethod
    def from_x_radians(cls, theta: float) -> SO3:
        return SO3.exp(np.array([theta, 0.0, 0.0], dtype=np.float64))

    @classmethod
    def from_y_radians(cls, theta: float) -> SO3:
        return SO3.exp(np.array([0.0, theta, 0.0], dtype=np.float64))

    @classmethod
    def from_z_radians(cls, theta: float) -> SO3:
        return SO3.exp(np.array([0.0, 0.0, theta], dtype=np.float64))

    @classmethod
    def from_rpy_radians(
        cls,
        roll: float,
        pitch: float,
        yaw: float,
    ) -> SO3:
        return (
            SO3.from_z_radians(yaw)
            @ SO3.from_y_radians(pitch)
            @ SO3.from_x_radians(roll)
        )

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> SO3:
        assert matrix.shape == (SO3.matrix_dim, SO3.matrix_dim)
        wxyz = np.empty(SO3.parameters_dim, dtype=np.float64)
        mujoco.mju_mat2Quat(wxyz, matrix.ravel())
        # NOTE mju_mat2Quat normalizes the quaternion.
        return SO3(wxyz=wxyz)

    @classmethod
    def identity(cls) -> SO3:
        return SO3(wxyz=_IDENTITIY_WXYZ)

    @classmethod
    def sample_uniform(cls) -> SO3:
        # Ref: https://lavalle.pl/planning/node198.html
        u1, u2, u3 = np.random.uniform(
            low=np.zeros(shape=(3,)),
            high=np.array([1.0, 2.0 * np.pi, 2.0 * np.pi]),
        )
        a = np.sqrt(1.0 - u1)
        b = np.sqrt(u1)
        wxyz = np.array(
            [
                a * np.sin(u2),
                a * np.cos(u2),
                b * np.sin(u3),
                b * np.cos(u3),
            ],
            dtype=np.float64,
        )
        return SO3(wxyz=wxyz)

    # Eq. 138.
    def as_matrix(self) -> np.ndarray:
        mat = np.empty(9, dtype=np.float64)
        mujoco.mju_quat2Mat(mat, self.wxyz)
        return mat.reshape(3, 3)

    def compute_roll_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arctan2(2 * (q0 * q1 + q2 * q3), 1 - 2 * (q1**2 + q2**2))

    def compute_pitch_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arcsin(2 * (q0 * q2 - q3 * q1))

    def compute_yaw_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arctan2(2 * (q0 * q3 + q1 * q2), 1 - 2 * (q2**2 + q3**2))

    def as_rpy_radians(self) -> RollPitchYaw:
        return RollPitchYaw(
            roll=self.compute_roll_radians(),
            pitch=self.compute_pitch_radians(),
            yaw=self.compute_yaw_radians(),
        )

    def inverse(self) -> SO3:
        conjugate_wxyz = np.empty(4)
        mujoco.mju_negQuat(conjugate_wxyz, self.wxyz)
        return SO3(wxyz=conjugate_wxyz)

    def normalize(self) -> SO3:
        normalized_wxyz = np.array(self.wxyz)
        mujoco.mju_normalize4(normalized_wxyz)
        return SO3(wxyz=normalized_wxyz)

    # Eq. 136.
    def apply(self, target: np.ndarray) -> np.ndarray:
        assert target.shape == (SO3.space_dim,)
        rotated_target = np.empty(SO3.space_dim, dtype=np.float64)
        mujoco.mju_rotVecQuat(rotated_target, target, self.wxyz)
        return rotated_target

    def multiply(self, other: SO3) -> SO3:
        res = np.empty(self.parameters_dim, dtype=np.float64)
        mujoco.mju_mulQuat(res, self.wxyz, other.wxyz)
        return SO3(wxyz=res)

    # Eq. 132.
    @classmethod
    def exp(cls, tangent: np.ndarray) -> SO3:
        axis = np.array(tangent)
        theta = mujoco.mju_normalize3(axis)
        wxyz = np.empty(4, dtype=np.float64)
        # NOTE mju_axisAngle2Quat does not normalize the quaternion but is guaranteed
        # to return a unit quaternion when axis is a unit vector. In our case,
        # mju_normalize3 ensures that axis is a unit vector.
        mujoco.mju_axisAngle2Quat(wxyz, axis, theta)
        return SO3(wxyz=wxyz)

    # Eq. 133.
    def log(self) -> np.ndarray:
        q = np.array(self.wxyz)
        q *= np.sign(q[0])
        w, v = q[0], q[1:]
        norm = mujoco.mju_normalize3(v)
        if norm < get_epsilon(v.dtype):
            return np.zeros_like(v)
        return 2 * np.arctan2(norm, w) * v

    # Eq. 139.
    def adjoint(self) -> np.ndarray:
        return self.as_matrix()

    # Jacobians.

    # Eqn. 145, 174.
    @classmethod
    def ljac(cls, other: np.ndarray) -> np.ndarray:
        theta = np.float64(mujoco.mju_norm3(other))
        t2 = theta * theta
        if theta < get_epsilon(theta.dtype):
            alpha = (1.0 / 2.0) * (
                1.0 - t2 / 12.0 * (1.0 - t2 / 30.0 * (1.0 - t2 / 56.0))
            )
            beta = (1.0 / 6.0) * (
                1.0 - t2 / 20.0 * (1.0 - t2 / 42.0 * (1.0 - t2 / 72.0))
            )
        else:
            t3 = t2 * theta
            alpha = (1 - np.cos(theta)) / t2
            beta = (theta - np.sin(theta)) / t3
        # ljac = eye(3) + alpha * skew_other + beta * (skew_other @ skew_other)
        ljac = np.empty((3, 3))
        # skew_other @ skew_other == outer(other) - inner(other) * eye(3)
        mujoco.mju_mulMatMat(ljac, other.reshape(3, 1), other.reshape(1, 3))
        inner_product = mujoco.mju_dot3(other, other)
        ljac[0, 0] -= inner_product
        ljac[1, 1] -= inner_product
        ljac[2, 2] -= inner_product
        ljac *= beta
        # + alpha * skew_other
        alpha_vec = alpha * other
        ljac[0, 1] += -alpha_vec[2]
        ljac[0, 2] += alpha_vec[1]
        ljac[1, 0] += alpha_vec[2]
        ljac[1, 2] += -alpha_vec[0]
        ljac[2, 0] += -alpha_vec[1]
        ljac[2, 1] += alpha_vec[0]
        # + eye(3)
        ljac[0, 0] += 1.0
        ljac[1, 1] += 1.0
        ljac[2, 2] += 1.0
        return ljac

    @classmethod
    def ljacinv(cls, other: np.ndarray) -> np.ndarray:
        theta = np.float64(mujoco.mju_norm3(other))
        t2 = theta * theta
        if theta < get_epsilon(theta.dtype):
            beta = (1.0 / 12.0) * (
                1.0 + t2 / 60.0 * (1.0 + t2 / 42.0 * (1.0 + t2 / 40.0))
            )
        else:
            beta = (1.0 / t2) * (
                1.0 - (theta * np.sin(theta) / (2.0 * (1.0 - np.cos(theta))))
            )
        # ljacinv = eye(3) - 0.5 * skew_other + beta * (skew_other @ skew_other)
        ljacinv = np.empty((3, 3))
        # skew_other @ skew_other == outer(other) - inner(other) * eye(3)
        mujoco.mju_mulMatMat(ljacinv, other.reshape(3, 1), other.reshape(1, 3))
        inner_product = mujoco.mju_dot3(other, other)
        ljacinv[0, 0] -= inner_product
        ljacinv[1, 1] -= inner_product
        ljacinv[2, 2] -= inner_product
        ljacinv *= beta
        # - 0.5 * skew_other
        alpha_vec = -0.5 * other
        ljacinv[0, 1] += -alpha_vec[2]
        ljacinv[0, 2] += alpha_vec[1]
        ljacinv[1, 0] += alpha_vec[2]
        ljacinv[1, 2] += -alpha_vec[0]
        ljacinv[2, 0] += -alpha_vec[1]
        ljacinv[2, 1] += alpha_vec[0]
        # + eye(3)
        ljacinv[0, 0] += 1.0
        ljacinv[1, 1] += 1.0
        ljacinv[2, 2] += 1.0
        return ljacinv



from functools import partial

import mujoco
import numpy as np

from . import keycodes


class TeleopMocap:
    """
    Class to handle keyboard input for teleoperation.
    The class provides methods to toggle teleoperation on/off,
    switch between manual and non-manual modes,
    adjust step sizes / speed for movement and rotation,
    and select movements based on key presses.
    """

    def __init__(self, data):
        self.on = False
        self.data = data
        self.reset_state()
        self.actions = {
            keycodes.KEY_N: self.toggle_manual,  # n: toggle non-manual mode
            keycodes.KEY_PERIOD: self.toggle_rotation,  # .: toggle rotation mode
            keycodes.KEY_8: self.toggle_mocap,  # 8: toggle mocap data
            keycodes.KEY_EQUAL: partial(self.toggle_speed, 1),  # =/+: increase speed
            keycodes.KEY_MINUS: partial(self.toggle_speed, -1),  # -: decrease speed
            keycodes.KEY_UP: partial(
                self.movement_select, keycodes.KEY_UP, 0, 1
            ),  # Up arrow
            keycodes.KEY_DOWN: partial(
                self.movement_select, keycodes.KEY_DOWN, 0, -1
            ),  # Down arrow
            keycodes.KEY_RIGHT: partial(
                self.movement_select, keycodes.KEY_RIGHT, 1, 1
            ),  # Right arrow
            keycodes.KEY_LEFT: partial(
                self.movement_select, keycodes.KEY_LEFT, 1, -1
            ),  # Left arrow
            keycodes.KEY_7: partial(self.movement_select, keycodes.KEY_7, 2, 1),  # 6
            keycodes.KEY_6: partial(self.movement_select, keycodes.KEY_6, 2, -1),  # 7
        }
        self.movements = {
            keycodes.KEY_UP: partial(self.movement_select, -1, 0, 1),  # Up arrow
            keycodes.KEY_DOWN: partial(self.movement_select, -1, 0, -1),  # Down arrow
            keycodes.KEY_RIGHT: partial(self.movement_select, -1, 1, 1),  # Right arrow
            keycodes.KEY_LEFT: partial(self.movement_select, -1, 1, -1),  # Left arrow
            keycodes.KEY_7: partial(self.movement_select, -1, 2, 1),  # 6
            keycodes.KEY_6: partial(self.movement_select, -1, 2, -1),  # 7
        }
        self.opposite_keys = {
            keycodes.KEY_UP: keycodes.KEY_DOWN,
            keycodes.KEY_DOWN: keycodes.KEY_UP,
            keycodes.KEY_RIGHT: keycodes.KEY_LEFT,
            keycodes.KEY_LEFT: keycodes.KEY_RIGHT,
            keycodes.KEY_7: keycodes.KEY_6,
            keycodes.KEY_6: keycodes.KEY_7,
        }

    def __call__(self, key):
        # Toggle teleop on/off
        if key == keycodes.KEY_9:  # 9
            self.toggle_on()
            return

        # Do nothing if teleop is off
        if not self.on:
            return

        if key in self.actions:
            self.actions[key]()

    def auto_key_move(self):
        """
        Automatically move the mocap body based on key presses.
        """

        if not self.on:
            return

        for key, action in self.movements.items():
            if self.keys[key]:
                action()

    def movement_select(self, key, axis, direction):
        """
        Select the movement direction based on the key pressed.
        """

        if not self.manual and key != -1:
            self.keys[key] = not self.keys[key]
            self.keys[self.opposite_keys[key]] = False
        elif not self.manual and key == -1:
            self.rot_or_trans(key, axis, direction)
        elif self.manual:
            self.rot_or_trans(key, axis, direction)

    def rot_or_trans(self, key, axis, direction):
        """
        Adjust the position or rotation of the mocap body based on rotation mode.
        """

        if self.rotation:
            self.adjust_rotation(key, axis, direction)
        else:
            self.adjust_position(key, axis, direction)

    def adjust_position(self, key, axis, direction):
        """
        Adjust the position of the mocap body in the specified direction
        based on the axis and step size.
        """

        q = self.data.mocap_quat[self.mocap_idx].copy()
        unit_vec = self.unit_vec_from_quat(q, axis)
        step_size = self.m_step_size if self.manual else self.nm_step_size
        self.data.mocap_pos[self.mocap_idx] += direction * step_size * unit_vec

    def adjust_rotation(self, key, axis, direction):
        """
        Adjust the rotation of the mocap body in the specified direction
        based on the axis and step size.
        """

        q = self.data.mocap_quat[self.mocap_idx].copy()
        unit_vec = self.unit_vec_from_quat(q, axis)

        # Rotate the quaternion by the specified angle around the axis.
        quat_rot = np.zeros(shape=(4,), dtype=np.float64)
        result = np.zeros(shape=(4,), dtype=np.float64)
        step_size = self.m_rotation_step if self.manual else self.nm_rotation_step
        angle = direction * step_size
        angle_rad = np.deg2rad(angle)
        unit_vec = unit_vec / np.linalg.norm(unit_vec)
        mujoco.mju_axisAngle2Quat(quat_rot, unit_vec, angle_rad)
        mujoco.mju_mulQuat(result, quat_rot, q)

        self.data.mocap_quat[self.mocap_idx] = result

    def unit_vec_from_quat(self, q, axis):
        """
        Compute the unit vector corresponding to the specified axis
        from the given quaternion.
        """

        rot = np.zeros(shape=(9,), dtype=np.float64)
        mujoco.mju_quat2Mat(rot, q)
        rot = rot.reshape((3, 3))
        unit_vec = rot[:, axis]

        return unit_vec

    def toggle_on(self):
        self.on = not self.on
        state = "On" if self.on else "Off"
        print(f"Keyboard Teleoperation toggled: {state}!")
        self.reset_state()
        print()

    def toggle_manual(self):
        self.manual = not self.manual
        manual_state = "On" if self.manual else "Off"
        print(f"Manual mode toggled: {manual_state}!")
        self.reset_keys()
        print()

    def toggle_rotation(self):
        self.rotation = not self.rotation
        state = "On" if self.rotation else "Off"
        print(f"Rotation mode toggled: {state}!")
        self.reset_keys()
        print()

    def toggle_speed(self, direction):
        factor = 1.10 if direction == 1 else 0.9
        if self.manual:
            if self.rotation:
                self.m_rotation_step *= factor
            else:
                self.m_step_size *= factor
        else:
            if self.rotation:
                self.nm_rotation_step *= factor
            else:
                self.nm_step_size *= factor

        output = "Manual" if self.manual else "Non-manual"
        mode = "Rotation" if self.rotation else "Translation"
        if self.manual:
            step_size = self.m_rotation_step if self.rotation else self.m_step_size
        else:
            step_size = self.nm_rotation_step if self.rotation else self.nm_step_size
        print(f"{output} {mode} step size: {step_size:.8f}")

    def toggle_mocap(self):
        self.mocap_idx = (self.mocap_idx + 1) % self.data.mocap_pos.shape[
            0
        ]  # cycle through mocap data
        print(f"Current mocap index: {self.mocap_idx}")

    def reset_keys(self):
        self.keys = {
            keycodes.KEY_UP: False,
            keycodes.KEY_DOWN: False,
            keycodes.KEY_RIGHT: False,
            keycodes.KEY_LEFT: False,
            keycodes.KEY_7: False,
            keycodes.KEY_6: False,
        }

    def reset_step_size(self):
        self.m_step_size = 0.01  # manual step size
        self.m_rotation_step = 10  # manual rotation step
        self.nm_step_size = 1e-4  # non-manual step size
        self.nm_rotation_step = 5e-2  # non-manual rotation step
        print("Step sizes have been reset!")

    def reset_state(self):
        self.reset_keys()
        self.reset_step_size()
        self.manual = True
        self.rotation = False
        self.mocap_idx = 0
        str = f"States have been reset: \n \
        - Manual mode: {self.manual} \
        - Rotation mode: {self.rotation} \n \
        - Mocap index: {self.mocap_idx}"

        print(str)
